# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-function-docstring,invalid-name

from __future__ import annotations

import contextlib
import math
import os
from typing import Literal

from nvitop.api import (
    NA,
    GiB,
    NaType,
    Snapshot,
    bytes2human,
    colored,
    set_color,
    timedelta2human,
    ttl_cache,
)
from nvitop.api.host import WINDOWS as IS_WINDOWS
from nvitop.api.host import WINDOWS_SUBSYSTEM_FOR_LINUX
from nvitop.tui.library.host import getuser, hostname
from nvitop.tui.library.widestring import WideString


__all__ = [
    'HOSTNAME',
    'IS_SUPERUSER',
    'IS_WINDOWS',
    'IS_WINDOWS_SUBSYSTEM_FOR_LINUX',
    'IS_WSL',
    'LARGE_INTEGER',
    'NA',
    'USERNAME',
    'USER_CONTEXT',
    'GiB',
    'NaType',
    'Snapshot',
    'bytes2human',
    'colored',
    'cut_string',
    'make_bar_chart',
    'set_color',
    'timedelta2human',
    'ttl_cache',
]


USERNAME: str = getuser()

IS_SUPERUSER: bool = False
with contextlib.suppress(AttributeError, OSError):
    if IS_WINDOWS:
        import ctypes

        IS_SUPERUSER = bool(ctypes.windll.shell32.IsUserAnAdmin())  # type: ignore[attr-defined]
    else:
        try:
            IS_SUPERUSER = os.geteuid() == 0
        except AttributeError:
            IS_SUPERUSER = os.getuid() == 0

HOSTNAME: str = hostname()
IS_WINDOWS_SUBSYSTEM_FOR_LINUX = IS_WSL = bool(WINDOWS_SUBSYSTEM_FOR_LINUX)
if IS_WSL:
    HOSTNAME = f'{HOSTNAME} (WSL)'

USER_CONTEXT: str = f'{USERNAME}@{HOSTNAME}'


LARGE_INTEGER: int = 65536


def cut_string(
    s: object,
    maxlen: int,
    padstr: str = '...',
    align: Literal['left', 'right'] = 'left',
) -> str:
    assert align in ('left', 'right')

    if not isinstance(s, str):
        s = str(s)
    s = WideString(s)
    padstr = WideString(padstr)

    if len(s) <= maxlen:
        return str(s)
    if len(padstr) >= maxlen:
        return str(padstr[:maxlen])
    if align == 'left':
        return str(s[: maxlen - len(padstr)] + padstr)
    return str(padstr + s[-(maxlen - len(padstr)) :])


# pylint: disable-next=too-many-arguments
def make_bar_chart(
    prefix: str,
    percent: float | str,
    width: int,
    *,
    extra_text: str = '',
    swap_text: bool = False,
    extra_blank: str = '',
) -> str:
    bar_chart = f'{prefix}: '
    if percent != NA and not (isinstance(percent, float) and not math.isfinite(percent)):
        if isinstance(percent, str) and percent.endswith('%'):
            percent = percent.replace('%', '')
            percent = float(percent) if '.' in percent else int(percent)
        percentage = max(0.0, min(float(percent) / 100.0, 1.0))
        quotient, remainder = divmod(
            max(1, round(8 * (width - len(bar_chart) - 4) * percentage)),
            8,
        )
        bar_chart += '█' * quotient
        if remainder > 0:
            bar_chart += ' ▏▎▍▌▋▊▉'[remainder]
        if isinstance(percent, float) and len(f'{bar_chart} {percent:.1f}%') <= width:
            text = f'{percent:.1f}%'
        else:
            text = f'{min(round(percent), 100):d}%'.replace('100%', 'MAX')  # type: ignore[arg-type]
    else:
        bar_chart += '░' * (width - len(bar_chart) - 4)
        text = 'N/A'
    if extra_text:
        if len(f'{bar_chart} {text} {extra_blank}{extra_text}') <= width:
            if swap_text:
                text, extra_text = extra_text, text
            return (
                f'{bar_chart} {text}'.ljust(width - len(extra_blank) - len(extra_text) - 1)
                + f' {extra_blank}{extra_text}'
            )
        if len(f'{bar_chart} {extra_text}') <= width and swap_text:
            return f'{bar_chart} {extra_text}'.ljust(width)
    return f'{bar_chart} {text}'.ljust(width)



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from nvitop.tui.library import Displayable, Selection


if TYPE_CHECKING:
    from nvitop.tui.screens.main import MainScreen
    from nvitop.tui.tui import TUI


__all__ = ['BasePanel', 'BaseSelectablePanel']


class BasePanel(Displayable):
    """Base class for all panels."""

    root: TUI
    parent: MainScreen

    NAME: ClassVar[str]
    SNAPSHOT_INTERVAL: ClassVar[float] = 0.5


class BaseSelectablePanel(BasePanel):
    """Base class for all selectable panels."""

    selection: Selection



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from nvitop.tui.library import host
from nvitop.tui.library.device import Device, MigDevice
from nvitop.tui.library.displayable import Displayable, DisplayableContainer
from nvitop.tui.library.history import BufferedHistoryGraph, HistoryGraph
from nvitop.tui.library.keybinding import (
    ALT_KEY,
    ANYKEY,
    PASSIVE_ACTION,
    QUANT_KEY,
    SPECIAL_KEYS,
    KeyBuffer,
    KeyMaps,
    normalize_keybinding,
)
from nvitop.tui.library.libcurses import libcurses, setlocale_utf8
from nvitop.tui.library.messagebox import MessageBox
from nvitop.tui.library.mouse import MouseEvent
from nvitop.tui.library.process import GpuProcess, HostProcess
from nvitop.tui.library.selection import Selection
from nvitop.tui.library.utils import (
    HOSTNAME,
    IS_SUPERUSER,
    IS_WINDOWS,
    IS_WINDOWS_SUBSYSTEM_FOR_LINUX,
    IS_WSL,
    LARGE_INTEGER,
    NA,
    USER_CONTEXT,
    USERNAME,
    GiB,
    NaType,
    Snapshot,
    bytes2human,
    colored,
    cut_string,
    make_bar_chart,
    set_color,
    timedelta2human,
    ttl_cache,
)
from nvitop.tui.library.widestring import WideString, wcslen


__all__ = [
    'ALT_KEY',
    'ANYKEY',
    'HOSTNAME',
    'IS_SUPERUSER',
    'IS_WINDOWS',
    'IS_WINDOWS_SUBSYSTEM_FOR_LINUX',
    'IS_WSL',
    'LARGE_INTEGER',
    'NA',
    'PASSIVE_ACTION',
    'QUANT_KEY',
    'SPECIAL_KEYS',
    'USERNAME',
    'USER_CONTEXT',
    'BufferedHistoryGraph',
    'Device',
    'Displayable',
    'DisplayableContainer',
    'GiB',
    'GpuProcess',
    'HistoryGraph',
    'HostProcess',
    'KeyBuffer',
    'KeyMaps',
    'MessageBox',
    'MigDevice',
    'MouseEvent',
    'NaType',
    'Selection',
    'Snapshot',
    'WideString',
    'bytes2human',
    'colored',
    'cut_string',
    'host',
    'libcurses',
    'make_bar_chart',
    'normalize_keybinding',
    'set_color',
    'setlocale_utf8',
    'timedelta2human',
    'ttl_cache',
    'wcslen',
]



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2025 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Shortcuts for package ``psutil``.

``psutil`` is a cross-platform library for retrieving information on running processes and system
utilization (CPU, memory, disks, network, sensors) in Python.
"""

from __future__ import annotations

import os as _os

import psutil as _psutil
from psutil import *  # noqa: F403 # pylint: disable=wildcard-import,unused-wildcard-import,redefined-builtin
from psutil import (  # noqa: F401
    LINUX,
    MACOS,
    POSIX,
    WINDOWS,
    AccessDenied,
    Error,
    NoSuchProcess,
    Process,
    ZombieProcess,
    boot_time,
    cpu_percent,
    pids,
    swap_memory,
    virtual_memory,
)
from psutil import Error as PsutilError  # pylint: disable=reimported


__all__ = [
    'WINDOWS_SUBSYSTEM_FOR_LINUX',
    'WSL',
    'PsutilError',
    'getuser',
    'hostname',
    'load_average',
    'memory_percent',
    'ppid_map',
    'reverse_ppid_map',
    'swap_percent',
    'uptime',
]
__all__ += [name for name in _psutil.__all__ if not name.startswith('_') and name != 'Error']


del Error  # renamed to PsutilError


def getuser() -> str:
    """Get the current username from the environment or password database."""
    import getpass  # pylint: disable=import-outside-toplevel

    try:
        return getpass.getuser()
    except (ModuleNotFoundError, OSError):
        return _os.getlogin()


def hostname() -> str:
    """Get the hostname of the machine."""
    import platform  # pylint: disable=import-outside-toplevel

    return platform.node()


if hasattr(_psutil, 'getloadavg'):

    def load_average() -> tuple[float, float, float]:
        """Get the system load average."""
        return _psutil.getloadavg()

else:

    def load_average() -> None:  # type: ignore[misc]
        """Get the system load average."""
        return


def uptime() -> float:
    """Get the system uptime."""
    import time as _time  # pylint: disable=import-outside-toplevel

    return _time.time() - boot_time()


def memory_percent() -> float:
    """The percentage usage of virtual memory, calculated as ``(total - available) / total * 100``."""
    return virtual_memory().percent


def swap_percent() -> float:
    """The percentage usage of virtual memory, calculated as ``used / total * 100``."""
    return swap_memory().percent


def ppid_map() -> dict[int, int]:
    """Obtain a ``{pid: ppid, ...}`` dict for all running processes in one shot."""
    ret = {}
    for pid in pids():
        try:
            ret[pid] = Process(pid).ppid()
        except (NoSuchProcess, ZombieProcess):  # noqa: PERF203
            pass
    return ret


try:
    from psutil import _ppid_map as ppid_map  # type: ignore[no-redef] # noqa: F811,RUF100
except ImportError:
    pass


def reverse_ppid_map() -> dict[int, list[int]]:
    """Obtain a ``{ppid: [pid, ...], ...}`` dict for all running processes in one shot."""
    from collections import defaultdict  # pylint: disable=import-outside-toplevel

    ret = defaultdict(list)
    for pid, ppid in ppid_map().items():
        ret[ppid].append(pid)

    return ret


# pylint: disable=invalid-name
if LINUX:
    WSL = _os.getenv('WSL_DISTRO_NAME', default=None)
    if WSL is not None and WSL == '':
        WSL = 'WSL'
else:
    WSL = None
WINDOWS_SUBSYSTEM_FOR_LINUX = WSL
"""The Linux distribution name of the Windows Subsystem for Linux."""
# pylint: enable=invalid-name



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from nvitop.tui.screens.main.panels.base import BasePanel, BaseSelectablePanel
from nvitop.tui.screens.main.panels.device import DevicePanel
from nvitop.tui.screens.main.panels.host import HostPanel
from nvitop.tui.screens.main.panels.process import OrderName, ProcessPanel


__all__ = [
    'BasePanel',
    'BaseSelectablePanel',
    'DevicePanel',
    'HostPanel',
    'OrderName',
    'ProcessPanel',
]



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# This file is originally part of ranger, the console file manager. https://github.com/ranger/ranger
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring

from __future__ import annotations

from typing import TYPE_CHECKING, Literal
from unicodedata import east_asian_width


if TYPE_CHECKING:
    from typing_extensions import Self  # Python 3.11+


__all__ = ['WideString', 'wcslen']


ASCIIONLY: frozenset[str] = frozenset(map(chr, range(1, 128)))
NARROW: Literal[1] = 1
WIDE: Literal[2] = 2
WIDE_SYMBOLS: frozenset[str] = frozenset('WF')


def utf_char_width(string: str) -> Literal[1, 2]:
    """Return the width of a single character."""
    if east_asian_width(string) in WIDE_SYMBOLS:
        return WIDE
    return NARROW


def string_to_charlist(string: str) -> list[str]:
    """Return a list of characters with extra empty strings after wide chars."""
    if ASCIIONLY.issuperset(string):
        return list(string)
    result = []
    for char in string:
        result.append(char)
        if east_asian_width(char) in WIDE_SYMBOLS:
            result.append('')
    return result


def wcslen(string: str | WideString) -> int:
    # pylint: disable=wrong-spelling-in-docstring
    """Return the length of a string with wide chars.

    >>> wcslen('poo')
    3
    >>> wcslen('十百千万')
    8
    >>> wcslen('a十')
    3
    """
    return len(WideString(string))


class WideString:  # pylint: disable=wrong-spelling-in-docstring
    def __init__(self, string: str | WideString = '', chars: list[str] | None = None) -> None:
        self.string: str = str(string)
        self.chars: list[str] = string_to_charlist(self.string) if chars is None else chars

    def __add__(self, other: object) -> WideString:
        """
        >>> (WideString('a') + WideString('b')).string
        'ab'
        >>> (WideString('a') + WideString('b')).chars
        ['a', 'b']
        >>> (WideString('afd') + 'bc').chars
        ['a', 'f', 'd', 'b', 'c']
        """
        if isinstance(other, str):
            return WideString(self.string + other)
        if isinstance(other, WideString):
            return WideString(self.string + other.string, self.chars + other.chars)
        return NotImplemented

    def __radd__(self, other: object) -> WideString:
        """
        >>> ('bc' + WideString('afd')).chars
        ['b', 'c', 'a', 'f', 'd']
        """
        if isinstance(other, str):
            return WideString(other + self.string)
        if isinstance(other, WideString):
            return WideString(other.string + self.string, other.chars + self.chars)
        return NotImplemented

    def __iadd__(self, other: object) -> Self:
        new = self + other
        self.string = new.string
        self.chars = new.chars
        return self

    def __str__(self) -> str:
        return self.string

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.string!r}>'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (str, WideString)):
            raise TypeError
        return str(self) == str(other)

    def __hash__(self) -> int:
        return hash(self.string)

    def __getitem__(self, item: int | slice) -> WideString:
        """
        >>> WideString('asdf')[2]
        <WideString 'd'>
        >>> WideString('……')[0]
        <WideString '…'>
        >>> WideString('……')[1]
        <WideString '…'>
        >>> WideString('asdf')[1:3]
        <WideString 'sd'>
        >>> WideString('asdf')[1:-100]
        <WideString ''>
        >>> WideString('十百千万')[2:4]
        <WideString '百'>
        >>> WideString('十百千万')[2:5]
        <WideString '百 '>
        >>> WideString('十ab千万')[2:5]
        <WideString 'ab '>
        >>> WideString('十百千万')[1:5]
        <WideString ' 百 '>
        >>> WideString('十百千万')[:]
        <WideString '十百千万'>
        >>> WideString('a十')[0:3]
        <WideString 'a十'>
        >>> WideString('a十')[0:2]
        <WideString 'a '>
        >>> WideString('a十')[0:1]
        <WideString 'a'>
        """
        if isinstance(item, slice):
            assert item.step is None or item.step == 1
            start, stop = item.start, item.stop
        else:
            assert isinstance(item, int)
            start, stop = item, item + 1

        length = len(self)

        if stop is None or stop > length:
            stop = length
        if stop < 0:
            stop = max(0, length + stop)
        if start is None:
            start = 0
        if start < 0:
            start = max(0, length + start)
        if start >= length or start >= stop:
            return WideString('')
        if stop < length and self.chars[stop] == '':
            if self.chars[start] == '':
                return WideString(' ' + ''.join(self.chars[start : stop - 1]) + ' ')
            return WideString(''.join(self.chars[start : stop - 1]) + ' ')
        if self.chars[start] == '':
            return WideString(' ' + ''.join(self.chars[start : stop - 1]))
        return WideString(''.join(self.chars[start:stop]))

    def __len__(self) -> int:
        """
        >>> len(WideString('poo'))
        3
        >>> len(WideString('十百千万'))
        8
        """
        return len(self.chars)

    def ljust(self, width: int, fillchar: str = ' ') -> WideString:
        """
        >>> WideString('poo').ljust(2)
        <WideString 'poo'>
        >>> WideString('poo').ljust(5)
        <WideString 'poo  '>
        >>> WideString('十百千万').ljust(10)
        <WideString '十百千万  '>
        """
        if width > len(self):
            return WideString(self.string + fillchar * width)[:width]
        return self

    def rjust(self, width: int, fillchar: str = ' ') -> WideString:
        """
        >>> WideString('poo').rjust(2)
        <WideString 'poo'>
        >>> WideString('poo').rjust(5)
        <WideString '  poo'>
        >>> WideString('十百千万').rjust(10)
        <WideString '  十百千万'>
        """
        if width > len(self):
            return WideString(fillchar * width + self.string)[-width:]
        return self

    def center(self, width: int, fillchar: str = ' ') -> WideString:
        """
        >>> WideString('poo').center(2)
        <WideString 'poo'>
        >>> WideString('poo').center(5)
        <WideString ' poo '>
        >>> WideString('十百千万').center(10)
        <WideString ' 十百千万 '>
        """
        if width > len(self):
            left_width = (width - len(self)) // 2
            right_width = width - left_width
            return WideString(fillchar * left_width + self.string + fillchar * right_width)[:width]
        return self

    def strip(self, chars: str | None = None) -> WideString:
        """
        >>> WideString('  poo  ').strip()
        <WideString 'poo'>
        >>> WideString('  十百千万  ').strip()
        <WideString '十百千万'>
        """
        return WideString(self.string.strip(chars))

    def lstrip(self, chars: str | None = None) -> WideString:
        """
        >>> WideString('  poo  ').lstrip()
        <WideString 'poo  '>
        >>> WideString('  十百千万  ').lstrip()
        <WideString '十百千万  '>
        """
        return WideString(self.string.lstrip(chars))

    def rstrip(self, chars: str | None = None) -> WideString:
        """
        >>> WideString('  poo  ').rstrip()
        <WideString '  poo'>
        >>> WideString('  十百千万  ').rstrip()
        <WideString '  十百千万'>
        """
        return WideString(self.string.rstrip(chars))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from nvitop.tui.screens.base import BaseScreen, BaseSelectableScreen
from nvitop.tui.screens.environ import EnvironScreen
from nvitop.tui.screens.help import HelpScreen
from nvitop.tui.screens.main import BreakLoop, MainScreen
from nvitop.tui.screens.metrics import ProcessMetricsScreen
from nvitop.tui.screens.treeview import TreeViewScreen


__all__ = [
    'BaseScreen',
    'BaseSelectableScreen',
    'BreakLoop',
    'EnvironScreen',
    'HelpScreen',
    'MainScreen',
    'ProcessMetricsScreen',
    'TreeViewScreen',
]



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from nvitop.tui.library import (
    IS_SUPERUSER,
    USERNAME,
    Device,
    colored,
    libcurses,
    set_color,
    setlocale_utf8,
)
from nvitop.tui.tui import TUI


__all__ = [
    'IS_SUPERUSER',
    'TUI',
    'USERNAME',
    'Device',
    'colored',
    'libcurses',
    'set_color',
    'setlocale_utf8',
]



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

from __future__ import annotations

import signal
import time
from typing import TYPE_CHECKING
from weakref import WeakValueDictionary

from nvitop.api import NA, Snapshot
from nvitop.tui.library import host
from nvitop.tui.library.utils import IS_SUPERUSER, IS_WINDOWS, LARGE_INTEGER, USERNAME


if TYPE_CHECKING:
    from collections.abc import Callable

    from nvitop.tui.library.displayable import Displayable
    from nvitop.tui.library.process import GpuProcess, HostProcess


__all__ = ['Selection']


class Selection:  # pylint: disable=too-many-instance-attributes
    def __init__(self, displayable: Displayable) -> None:
        self.tagged: WeakValueDictionary[int, GpuProcess | HostProcess] = WeakValueDictionary()
        self.displayable: Displayable = displayable
        self.index: int | None = None
        self.within_window: bool = True
        self._process: GpuProcess | HostProcess | None = None
        self._username: str | None = None
        self._ident: tuple | None = None

    @property
    def identity(self) -> tuple:
        if self._ident is None:
            self._ident = self.process._ident  # type: ignore[union-attr] # pylint: disable=protected-access
        return self._ident

    @property
    def process(self) -> GpuProcess | HostProcess | None:
        return self._process

    @process.setter
    def process(self, process: Snapshot | GpuProcess | HostProcess | None) -> None:
        if isinstance(process, Snapshot):
            process = process.real
        self._process = process  # type: ignore[assignment]
        self._ident = None

    @property
    def pid(self) -> int | None:
        try:
            return self.identity[0]
        except TypeError:
            return None

    @property
    def username(self) -> str | None:
        if self._username is None:
            try:
                self._username = self.process.username()  # type: ignore[union-attr]
            except host.PsutilError:
                self._username = NA
        return self._username

    def move(self, direction: int = 0) -> None:
        if direction == 0:
            return

        processes = self.displayable.snapshots  # type: ignore[attr-defined]
        old_index = self.index

        if len(processes) > 0:
            if not self.is_set():
                if abs(direction) < LARGE_INTEGER:
                    self.index = 0 if direction > 0 else len(processes) - 1
                else:
                    self.index = len(processes) - 1 if direction > 0 else 0
            else:
                self.index = min(max(0, self.index + direction), len(processes) - 1)  # type: ignore[operator]
            self.process = processes[self.index]

            if old_index is not None:
                direction -= self.index - old_index
            else:
                direction = 0

            if direction != 0 and self.displayable.NAME == 'process':  # type: ignore[attr-defined]
                self.displayable.parent.move(direction)  # type: ignore[union-attr]
        else:
            self.clear()

    def owned(self) -> bool:
        if not self.is_set():
            return False
        return IS_SUPERUSER or self.username == USERNAME

    def tag(self) -> None:
        if self.is_set():
            try:
                del self.tagged[self.pid]  # type: ignore[arg-type]
            except KeyError:
                self.tagged[self.pid] = self.process  # type: ignore[index,assignment]

    def processes(self) -> tuple[GpuProcess | HostProcess, ...]:
        if len(self.tagged) > 0:
            return tuple(sorted(self.tagged.values(), key=lambda p: p.pid))
        if self.owned() and self.within_window:
            return (self.process,)  # type: ignore[return-value]
        return ()

    def foreach(self, func: Callable[[GpuProcess | HostProcess], None]) -> None:
        flag = False
        for process in self.processes():
            try:
                func(process)
            except host.PsutilError:  # noqa: PERF203
                pass
            else:
                flag = True

        if flag:
            time.sleep(0.25)
        self.clear()

    def send_signal(self, sig: int) -> None:
        self.foreach(lambda process: process.send_signal(sig))

    def interrupt(self) -> None:
        try:
            # pylint: disable-next=no-member
            self.send_signal(signal.SIGINT if not IS_WINDOWS else signal.CTRL_C_EVENT)  # type: ignore[attr-defined]
        except SystemError:
            pass

    def terminate(self) -> None:
        self.foreach(lambda process: process.terminate())

    def kill(self) -> None:
        self.foreach(lambda process: process.kill())

    def reset(self) -> None:
        self.index = None
        self.within_window = True
        self._process = None
        self._username = None
        self._ident = None

    def clear(self) -> None:
        self.tagged.clear()
        self.reset()

    def is_set(self) -> bool:
        return self.process is not None

    def is_same(self, process: Snapshot | GpuProcess | HostProcess) -> bool:
        if isinstance(process, Snapshot):
            process = process.real

        try:
            return self.identity == process._ident  # pylint: disable=protected-access
        except (AttributeError, TypeError):
            pass

        return False

    def is_same_on_host(self, process: Snapshot | GpuProcess | HostProcess) -> bool:
        if isinstance(process, Snapshot):
            process = process.real

        try:
            return self.identity[:2] == process._ident[:2]  # pylint: disable=protected-access
        except (AttributeError, TypeError):
            pass

        return False

    def is_tagged(self, process: Snapshot | GpuProcess | HostProcess) -> bool:
        return process.pid in self.tagged



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

from __future__ import annotations

import colorsys
import contextlib
import curses
import locale
import os
import signal
from typing import TYPE_CHECKING, Any, ClassVar, Tuple, Union

from nvitop.tui.library.history import GRAPH_SYMBOLS


if TYPE_CHECKING:
    from collections.abc import Generator
    from typing_extensions import TypeAlias  # Python 3.10+


__all__ = ['CursesShortcuts', 'libcurses', 'setlocale_utf8']


# pylint: disable=invalid-name
LIGHT_THEME: bool = False
DEFAULT_FOREGROUND: int = curses.COLOR_WHITE
DEFAULT_BACKGROUND: int = curses.COLOR_BLACK
COLOR_PAIRS: dict[tuple[int, int], int] = {}
TRUE_COLORS: dict[str | tuple[int, int, int], int] = {
    'black': 0,
    'red': 1,
    'green': 2,
    'yellow': 3,
    'blue': 4,
    'magenta': 5,
    'cyan': 6,
    'white': 7,
    'bright black': 8,
    'bright red': 9,
    'bright green': 10,
    'bright yellow': 11,
    'bright blue': 12,
    'bright magenta': 13,
    'bright cyan': 14,
    'bright white': 15,
    **{f'preserved {i:02d}': i for i in range(16, 64)},
}
BASE_ATTR: int = 0
# pylint: enable=invalid-name


def _init_color_theme(light_theme: bool = False) -> None:
    """Set the default fg/bg colors."""
    global LIGHT_THEME, DEFAULT_FOREGROUND, DEFAULT_BACKGROUND  # pylint: disable=global-statement

    LIGHT_THEME = light_theme
    if LIGHT_THEME:
        DEFAULT_FOREGROUND = curses.COLOR_BLACK
        DEFAULT_BACKGROUND = curses.COLOR_WHITE
    else:
        DEFAULT_FOREGROUND = curses.COLOR_WHITE
        DEFAULT_BACKGROUND = curses.COLOR_BLACK


def _colormap(x: float, levels: int = 160) -> tuple[int, int, int]:
    # pylint: disable=invalid-name
    h = 0.5 * (1.0 - x) - 0.15
    h = (round(h * levels) / levels) % 1.0
    r, g, b = colorsys.hsv_to_rgb(h, 0.7, 0.8)
    return (round(1000.0 * r), round(1000.0 * g), round(1000.0 * b))


def _get_true_color(rgb: tuple[int, int, int]) -> int:
    if rgb not in TRUE_COLORS:
        try:
            curses.init_color(len(TRUE_COLORS), *rgb)
        except curses.error:
            return -1
        TRUE_COLORS[rgb] = len(TRUE_COLORS)
    return TRUE_COLORS[rgb]


Color: TypeAlias = Union[str, int, float, Tuple[int, int, int]]


def _get_color(fg: Color, bg: Color) -> int:
    """Return the curses color pair for the given fg/bg combination."""
    global COLOR_PAIRS  # pylint: disable=global-statement,global-variable-not-assigned

    if isinstance(fg, str):
        fg = getattr(curses, f'COLOR_{fg.upper()}', -1)
    elif isinstance(fg, tuple):
        fg = _get_true_color(fg)
    elif isinstance(fg, float):
        fg = _get_true_color(_colormap(fg))
    if isinstance(bg, str):
        bg = getattr(curses, f'COLOR_{bg.upper()}', -1)
    elif isinstance(bg, tuple):
        bg = _get_true_color(bg)
    elif isinstance(bg, float):
        bg = _get_true_color(_colormap(bg))

    key = (fg, bg)
    if key not in COLOR_PAIRS:
        new_id = len(COLOR_PAIRS) + 1
        try:
            curses.init_pair(new_id, fg, bg)
        except curses.error:
            # If curses.use_default_colors() failed during the initialization
            # of curses, then using -1 as fg or bg will fail as well, which
            # we need to handle with fallback-defaults:
            if fg == -1:  # -1 is the "default" color
                fg = DEFAULT_FOREGROUND
            if bg == -1:  # -1 is the "default" color
                bg = DEFAULT_BACKGROUND

            try:
                curses.init_pair(new_id, fg, bg)
            except curses.error:
                # If this fails too, colors are probably not supported
                pass
        COLOR_PAIRS[key] = new_id

    return COLOR_PAIRS[key]


def setlocale_utf8() -> bool:
    for code in ('C.UTF-8', 'en_US.UTF-8', '', 'C'):
        try:
            code = locale.setlocale(locale.LC_ALL, code)
        except locale.Error:  # noqa: PERF203
            continue
        else:
            if 'utf8' in code.lower() or 'utf-8' in code.lower():
                return True

    return False


@contextlib.contextmanager
def libcurses(colorful: bool = False, light_theme: bool = False) -> Generator[curses.window]:
    os.environ.setdefault('ESCDELAY', '25')
    setlocale_utf8()

    win = curses.initscr()
    win.nodelay(True)
    win.leaveok(True)
    win.keypad(True)

    curses.noecho()
    curses.cbreak()
    curses.curs_set(False)
    curses.mousemask(curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION)
    curses.mouseinterval(0)
    curses.ungetmouse(0, 0, 0, 0, 0)

    _init_color_theme(light_theme)

    curses.start_color()
    try:
        curses.use_default_colors()
    except curses.error:
        pass

    if colorful:
        try:
            CursesShortcuts.TERM_256COLOR = curses.COLORS >= 256
        except AttributeError:
            pass

    # Push a Ctrl+C (ascii value 3) to the curses getch stack
    def interrupt_handler(*_: Any) -> None:  # pylint: disable=unused-argument
        curses.ungetch(3)

    # Simulate a ^C press in curses when an interrupt is caught
    signal.signal(signal.SIGINT, interrupt_handler)

    try:
        yield win
    finally:
        curses.endwin()


class CursesShortcuts:
    """This class defines shortcuts to facilitate operations with curses.

    color(*keys) -- sets the color associated with the keys from
        the current colorscheme.
    color_at(y, x, width, *keys) -- sets the color at the given position
    color_reset() -- resets the color to the default
    addstr(*args) -- failsafe version of self.win.addstr(*args)
    """

    ASCII_TRANSTABLE: ClassVar[dict[int, int]] = str.maketrans(
        '═─╴╒╤╕╪╘╧╛┌┬┐┼└┴┘│╞╡├┤▏▎▍▌▋▊▉█░▲▼␤' + GRAPH_SYMBOLS,
        '=--++++++++++++++||||||||||||||^v?' + '=' * len(GRAPH_SYMBOLS),
    )
    TERM_256COLOR: ClassVar[bool] = False

    def __init__(self) -> None:
        self.win: curses.window | None = None
        self.no_unicode: bool = False

    def addstr(self, *args: str | int | Color, **kwargs: str | int | Color) -> None:
        if self.no_unicode:
            args = [  # type: ignore[assignment]
                arg.translate(self.ASCII_TRANSTABLE) if isinstance(arg, str) else arg
                for arg in args
            ]

        assert self.win is not None
        try:
            self.win.addstr(*args, **kwargs)  # type: ignore[arg-type]
        except curses.error:
            pass

    def addnstr(self, *args: str | int | Color, **kwargs: str | int | Color) -> None:
        if self.no_unicode:
            args = [  # type: ignore[assignment]
                arg.translate(self.ASCII_TRANSTABLE) if isinstance(arg, str) else arg
                for arg in args
            ]

        assert self.win is not None
        try:
            self.win.addnstr(*args, **kwargs)  # type: ignore[arg-type]
        except curses.error:
            pass

    def addch(self, *args: str | int | Color, **kwargs: str | int | Color) -> None:
        if self.no_unicode:
            args = [  # type: ignore[assignment]
                arg.translate(self.ASCII_TRANSTABLE) if isinstance(arg, str) else arg
                for arg in args
            ]

        assert self.win is not None
        try:
            self.win.addch(*args, **kwargs)  # type: ignore[arg-type]
        except curses.error:
            pass

    def color(self, fg: Color = -1, bg: Color = -1, attr: str | int = 0) -> int:
        """Change the colors from now on."""
        return self.set_fg_bg_attr(fg, bg, attr)

    def color_reset(self) -> int:
        """Change the colors to the default colors."""
        return self.color()

    def color_at(
        self,
        y: int,
        x: int,
        width: int,
        *args: str | int | Color,
        **kwargs: str | int | Color,
    ) -> None:
        """Change the colors at the specified position."""
        assert self.win is not None
        try:
            self.win.chgat(y, x, width, self.get_fg_bg_attr(*args, **kwargs))  # type: ignore[arg-type]
        except curses.error:
            pass

    @staticmethod
    def get_fg_bg_attr(fg: Color = -1, bg: Color = -1, attr: str | int = 0) -> int:
        """Return the curses attribute for the given fg/bg/attr combination."""
        if fg == -1 and bg == -1 and attr == 0:
            return BASE_ATTR

        if isinstance(attr, str):
            attr_strings = map(str.strip, attr.split('|'))
            attr = 0
            for s in attr_strings:
                attr |= getattr(curses, f'A_{s.upper()}', 0)

        # Tweak for light themes
        if (
            LIGHT_THEME
            and attr & curses.A_REVERSE != 0
            and bg == -1
            and fg not in {DEFAULT_FOREGROUND, -1}
        ):
            bg = DEFAULT_FOREGROUND

        if fg == -1 and bg == -1:
            return attr | BASE_ATTR
        return curses.color_pair(_get_color(fg, bg)) | attr | BASE_ATTR

    def set_fg_bg_attr(self, fg: Color = -1, bg: Color = -1, attr: str | int = 0) -> int:
        assert self.win is not None
        try:
            attr = self.get_fg_bg_attr(fg, bg, attr)
            self.win.attrset(attr)
        except curses.error:
            return 0
        return attr

    def update_size(self, termsize: tuple[int, int] | None = None) -> tuple[int, int]:
        if termsize is not None:
            return termsize

        self.update_lines_cols()
        assert self.win is not None
        return self.win.getmaxyx()

    @staticmethod
    def update_lines_cols() -> None:
        curses.update_lines_cols()

    @staticmethod
    def beep() -> None:
        curses.beep()

    @staticmethod
    def flash() -> None:
        curses.flash()

    @staticmethod
    def set_base_attr(attr: str | int = 0) -> None:
        global BASE_ATTR  # pylint: disable=global-statement

        if isinstance(attr, str):
            attr_strings = map(str.strip, attr.split('|'))
            attr = 0
            for s in attr_strings:
                attr |= getattr(curses, f'A_{s.upper()}', 0)

        BASE_ATTR = attr



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# This file is originally part of ranger, the console file manager. https://github.com/ranger/ranger
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

from __future__ import annotations

import copy
import curses
import curses.ascii
import string
from collections import OrderedDict
from typing import TYPE_CHECKING, Callable, Dict, Tuple, Union


if TYPE_CHECKING:
    from collections.abc import Generator
    from typing_extensions import TypeAlias  # Python 3.10+


__all__ = [
    'ALT_KEY',
    'ANYKEY',
    'DIGITS',
    'PASSIVE_ACTION',
    'QUANT_KEY',
    'REVERSED_SPECIAL_KEYS',
    'SPECIAL_KEYS',
    'SPECIAL_KEYS',
    'SPECIAL_KEYS_UNCASED',
    'KeyBuffer',
    'KeyMaps',
    'construct_keybinding',
    'normalize_keybinding',
    'parse_keybinding',
]


IntKey: TypeAlias = Union[int, Tuple[int, ...]]


DIGITS: frozenset[int] = frozenset(map(ord, string.digits))

# Arbitrary numbers which are not used with curses.KEY_XYZ
ANYKEY: int = 9001
PASSIVE_ACTION: int = 9002
ALT_KEY: int = 9003
QUANT_KEY: int = 9004


NAMED_SPECIAL_KEYS: OrderedDict[str, int] = OrderedDict(
    [
        ('BS', curses.KEY_BACKSPACE),
        ('Backspace', curses.KEY_BACKSPACE),  # overrides <BS> in REVERSED_SPECIAL_KEYS
        ('Backspace2', curses.ascii.DEL),
        ('Delete', curses.KEY_DC),
        ('S-Delete', curses.KEY_SDC),
        ('Insert', curses.KEY_IC),
        ('CR', ord('\n')),
        ('Return', ord('\n')),
        ('Enter', ord('\n')),  # overrides <CR> and <Return> in REVERSED_SPECIAL_KEYS
        ('Space', ord(' ')),
        ('Escape', curses.ascii.ESC),
        ('Esc', curses.ascii.ESC),  # overrides <Escape> in REVERSED_SPECIAL_KEYS
        ('Down', curses.KEY_DOWN),
        ('Up', curses.KEY_UP),
        ('Left', curses.KEY_LEFT),
        ('Right', curses.KEY_RIGHT),
        ('PageDown', curses.KEY_NPAGE),
        ('PageUp', curses.KEY_PPAGE),
        ('Home', curses.KEY_HOME),
        ('End', curses.KEY_END),
        ('Tab', ord('\t')),
        ('S-Tab', curses.KEY_BTAB),
        ('lt', ord('<')),
        ('gt', ord('>')),
    ],
)
SPECIAL_KEYS: OrderedDict[str, IntKey]
SPECIAL_KEYS_UNCASED: dict[str, IntKey]
REVERSED_SPECIAL_KEYS: dict[IntKey, str]


def _uncase_special_key(key_string: str) -> str:
    """Uncase a special key.

    >>> _uncase_special_key('Esc')
    'esc'

    >>> _uncase_special_key('C-X')
    'c-x'
    >>> _uncase_special_key('C-x')
    'c-x'

    >>> _uncase_special_key('A-X')
    'a-X'
    >>> _uncase_special_key('A-x')
    'a-x'
    """
    uncased = key_string.lower()
    if len(uncased) == 3 and (uncased.startswith(('a-', 'm-'))):
        uncased = f'{uncased[0]}-{key_string[-1]}'
    return uncased


def _special_keys_init() -> None:
    global SPECIAL_KEYS, SPECIAL_KEYS_UNCASED, REVERSED_SPECIAL_KEYS  # pylint: disable=global-statement

    SPECIAL_KEYS = NAMED_SPECIAL_KEYS.copy()  # type: ignore[assignment]
    for key, int_value in NAMED_SPECIAL_KEYS.items():
        SPECIAL_KEYS[f'M-{key}'] = (ALT_KEY, int_value)
        SPECIAL_KEYS[f'A-{key}'] = (ALT_KEY, int_value)  # overrides <M-*> in REVERSED_SPECIAL_KEYS

    for char in 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_!{}[],./':
        SPECIAL_KEYS[f'M-{char}'] = (ALT_KEY, ord(char))
        SPECIAL_KEYS[f'A-{char}'] = (ALT_KEY, ord(char))  # overrides <M-*> in REVERSED_SPECIAL_KEYS

    # We will need to reorder the keys of SPECIAL_KEYS below.
    # For example, <C-j> will override <Enter> in REVERSE_SPECIAL_KEYS,
    # this makes construct_keybinding(parse_keybinding('<CR>')) == '<C-j>'
    for char in 'abcdefghijklmnopqrstuvwxyz_':
        SPECIAL_KEYS[f'C-{char}'] = ord(char) - 96
    SPECIAL_KEYS['C-Space'] = 0

    for n in range(64):
        SPECIAL_KEYS[f'F{n}'] = curses.KEY_F0 + n

    # Very special keys
    SPECIAL_KEYS.update(
        {
            'Alt': ALT_KEY,
            'any': ANYKEY,
            'bg': PASSIVE_ACTION,
            'allow_quantifiers': QUANT_KEY,
        },
    )

    # Reorder the keys of SPECIAL_KEYS
    for key in NAMED_SPECIAL_KEYS:
        SPECIAL_KEYS.move_to_end(key, last=True)

    SPECIAL_KEYS_UNCASED = {_uncase_special_key(k): v for k, v in SPECIAL_KEYS.items()}
    REVERSED_SPECIAL_KEYS = {v: k for k, v in SPECIAL_KEYS.items()}


_special_keys_init()
del _special_keys_init


def parse_keybinding(obj: IntKey | str) -> tuple[int, ...]:  # pylint: disable=too-many-branches
    r"""Translate a keybinding to a sequence of integers
    The letter case of special keys in the keybinding string will be ignored.

    >>> out = parse_keybinding('lol<CR>')
    >>> out
    (108, 111, 108, 10)
    >>> out == (ord('l'), ord('o'), ord('l'), ord('\n'))
    True

    >>> out = parse_keybinding('x<A-Left>')
    >>> out
    (120, 9003, 260)
    >>> out == (ord('x'), ALT_KEY, curses.KEY_LEFT)
    True
    """
    assert isinstance(obj, (tuple, int, str))

    def parse(obj: IntKey | str) -> Generator[int]:  # pylint: disable=too-many-branches
        if isinstance(obj, tuple):
            yield from obj
            return
        if isinstance(obj, int):
            yield obj
            return

        in_brackets = False
        bracket_content: list[str] = []
        for char in obj:
            if in_brackets:
                if char == '>':
                    in_brackets = False
                    key_string = ''.join(bracket_content)
                    try:
                        keys = SPECIAL_KEYS_UNCASED[_uncase_special_key(key_string)]
                        yield from keys  # type: ignore[misc]
                    except KeyError:
                        if key_string.isdigit():
                            yield int(key_string)
                        else:
                            yield ord('<')
                            for bracket_char in bracket_content:
                                yield ord(bracket_char)
                            yield ord('>')
                    except TypeError:
                        yield keys  # type: ignore[misc] # it was not tuple, just an int
                else:
                    bracket_content.append(char)
            elif char == '<':
                in_brackets = True
                bracket_content = []
            else:
                yield ord(char)
        if in_brackets:
            yield ord('<')
            for char in bracket_content:
                yield ord(char)

    return tuple(parse(obj))


def key_to_string(key: IntKey) -> str:
    if key in range(33, 127):
        return chr(key)  # type: ignore[arg-type]
    return f'<{REVERSED_SPECIAL_KEYS.get(key, key)}>'


def construct_keybinding(keys: IntKey) -> str:
    """Do the reverse of parse_keybinding.

    >>> construct_keybinding(parse_keybinding('lol<CR>'))
    'lol<Enter>'

    >>> construct_keybinding(parse_keybinding('x<A-Left>'))
    'x<A-Left>'

    >>> construct_keybinding(parse_keybinding('x<Alt><Left>'))
    'x<A-Left>'
    """
    keys = (keys,) if isinstance(keys, int) else tuple(keys)
    strings = []
    alt_key_on = False
    for key in keys:
        if key == ALT_KEY:
            alt_key_on = True
            continue
        if alt_key_on:
            try:
                strings.append(f'<{REVERSED_SPECIAL_KEYS[ALT_KEY, key]}>')
            except KeyError:
                strings.extend(map(key_to_string, (ALT_KEY, key)))
        else:
            strings.append(key_to_string(key))
        alt_key_on = False

    return ''.join(strings)


def normalize_keybinding(keybinding: str) -> str:
    """Normalize a keybinding to a string.

    >>> normalize_keybinding('lol<CR>')
    'lol<Enter>'

    >>> normalize_keybinding('x<A-Left>')
    'x<A-Left>'

    >>> normalize_keybinding('x<Alt><Left>')
    'x<A-Left>'
    """
    return construct_keybinding(parse_keybinding(keybinding))


KeyMapPointer: TypeAlias = Dict[int, Union['KeyMapPointer', Callable[[], None]]]


class KeyMaps(Dict[str, KeyMapPointer]):
    def __init__(self, keybuffer: KeyBuffer) -> None:
        super().__init__()
        self.keybuffer: KeyBuffer = keybuffer
        self.used_keymap: str | None = None

    def use_keymap(self, keymap_name: str) -> None:
        self.keybuffer.keymap = self.get(keymap_name, {})
        if self.used_keymap != keymap_name:
            self.used_keymap = keymap_name
            self.keybuffer.clear()

    def clear_keymap(self, keymap_name: str) -> KeyMapPointer:
        keymap = self[keymap_name] = {}
        if self.used_keymap == keymap_name:
            self.keybuffer.keymap = keymap
            self.keybuffer.clear()
        return keymap

    def _clean_input(self, context: str, keybinding: str) -> tuple[tuple[int, ...], KeyMapPointer]:
        try:
            pointer = self[context]
        except KeyError:
            self[context] = pointer = {}
        keybinding = keybinding.encode('utf-8').decode('latin-1')
        return parse_keybinding(keybinding), pointer

    def bind(self, context: str, keybinding: str, leaf: Callable[[], None]) -> None:
        keys, pointer = self._clean_input(context, keybinding)
        if not keys:
            return
        last_key = keys[-1]
        for key in keys[:-1]:
            if key in pointer and isinstance(pointer[key], dict):
                pointer = pointer[key]  # type: ignore[assignment]
            else:
                pointer = pointer[key] = {}
        pointer[last_key] = leaf

    def alias(self, context: str, source: str, target: str) -> None:
        clean_source, pointer = self._clean_input(context, source)
        if not source:
            return
        for key in clean_source:
            try:
                pointer = pointer[key]  # type: ignore[assignment]
            except KeyError as ex:  # noqa: PERF203
                raise KeyError(
                    f'Tried to copy the keybinding `{source}`, but it was not found.',
                ) from ex
        try:
            self.bind(context, target, copy.deepcopy(pointer))  # type: ignore[arg-type]
        except TypeError:
            self.bind(context, target, pointer)  # type: ignore[arg-type]

    def unbind(self, context: str, keybinding: str) -> None:
        keys, pointer = self._clean_input(context, keybinding)
        if not keys:
            return

        def unbind_traverse(pointer: KeyMapPointer, keys: list[int], pos: int = 0) -> None:
            if keys[pos] not in pointer:
                return
            if len(keys) > pos + 1 and isinstance(pointer, dict):
                unbind_traverse(pointer[keys[pos]], keys, pos=pos + 1)  # type: ignore[arg-type]
                if not pointer[keys[pos]]:
                    del pointer[keys[pos]]
            elif len(keys) == pos + 1:
                try:
                    del pointer[keys[pos]]
                except KeyError:
                    pass
                try:
                    keys.pop()
                except IndexError:
                    pass

        unbind_traverse(pointer, list(keys))


class KeyBuffer:  # pylint: disable=too-many-instance-attributes
    class QuantifierFinished:  # pylint: disable=too-few-public-methods
        pass

    QUANTIFIER_KEY_FINISHED = QuantifierFinished()

    del QuantifierFinished

    any_key: int = ANYKEY
    passive_key: int = PASSIVE_ACTION
    quantifier_key: int = QUANT_KEY
    excluded_from_anykey: frozenset[int] = frozenset({curses.ascii.ESC})

    def __init__(self, keymap: KeyMapPointer | None = None) -> None:
        self.keymap: KeyMapPointer | None = keymap
        self.keys: list[int] = []
        self.wildcards: list[int] = []
        self.pointer: KeyMapPointer | None = self.keymap
        self.result: Callable[[], None] | None = None
        self.quantifier: int | None = None
        self.finished_parsing_quantifier: bool = False
        self.finished_parsing: bool = False
        self.parse_error: bool = False

        if (
            self.keymap
            and self.quantifier_key in self.keymap
            and self.keymap[self.quantifier_key] is self.QUANTIFIER_KEY_FINISHED  # type: ignore[comparison-overlap]
        ):
            self.finished_parsing_quantifier = True

    def clear(self) -> None:
        self.__init__(self.keymap)  # type: ignore[misc] # pylint: disable=unnecessary-dunder-call

    def add(self, key: int) -> None:
        assert self.pointer is not None
        self.keys.append(key)
        self.result = None
        if not self.finished_parsing_quantifier and key in DIGITS:
            if self.quantifier is None:
                self.quantifier = 0
            self.quantifier = self.quantifier * 10 + key - 48  # (48 = ord('0'))
        else:
            self.finished_parsing_quantifier = True

            moved = True
            if key in self.pointer:
                self.pointer = self.pointer[key]  # type: ignore[assignment]
            elif self.any_key in self.pointer and key not in self.excluded_from_anykey:
                self.wildcards.append(key)
                self.pointer = self.pointer[self.any_key]  # type: ignore[assignment]
            else:
                moved = False

            if moved:
                if isinstance(self.pointer, dict):
                    if self.passive_key in self.pointer:
                        self.result = self.pointer[self.passive_key]  # type: ignore[assignment]
                else:
                    self.result = self.pointer
                    self.finished_parsing = True
            else:
                self.finished_parsing = True
                self.parse_error = True

    def __str__(self) -> str:
        return construct_keybinding(tuple(self.keys))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

from __future__ import annotations

import enum
from typing import Any, ClassVar, Literal

from nvitop.api import NA, NaType, Snapshot, libnvml, ttl_cache, utilization2string
from nvitop.api import MigDevice as MigDeviceBase
from nvitop.api import PhysicalDevice as DeviceBase
from nvitop.tui.library.process import GpuProcess, GpuProcessBase


__all__ = ['Device', 'MigDevice']


class LoadingIntensity(enum.IntEnum):
    LIGHT = 0
    MODERATE = 1
    HEAVY = 2

    def color(self) -> str:
        if self == LoadingIntensity.LIGHT:
            return 'green'
        if self == LoadingIntensity.MODERATE:
            return 'yellow'
        return 'red'


class Device(DeviceBase):  # pylint: disable=too-many-public-methods
    GPU_PROCESS_CLASS: ClassVar[type[GpuProcessBase]] = GpuProcess

    MEMORY_UTILIZATION_THRESHOLDS: ClassVar[tuple[int, int]] = (10, 80)
    GPU_UTILIZATION_THRESHOLDS: ClassVar[tuple[int, int]] = (10, 75)

    SNAPSHOT_KEYS: ClassVar[list[str]] = [
        'name',
        'bus_id',
        'memory_used',
        'memory_free',
        'memory_total',
        'memory_used_human',
        'memory_free_human',
        'memory_total_human',
        'memory_percent',
        'memory_usage',
        'gpu_utilization',
        'memory_utilization',
        'fan_speed',
        'temperature',
        'power_usage',
        'power_limit',
        'power_status',
        'display_active',
        'current_driver_model',
        'persistence_mode',
        'performance_state',
        'total_volatile_uncorrected_ecc_errors',
        'compute_mode',
        'mig_mode',
        'is_mig_device',
        'power_utilization',
        'memory_percent_string',
        'memory_utilization_string',
        'gpu_utilization_string',
        'fan_speed_string',
        'temperature_string',
        'memory_loading_intensity',
        'memory_display_color',
        'bandwidth_loading_intensity',
        'bandwidth_display_color',
        'gpu_loading_intensity',
        'gpu_display_color',
        'power_loading_intensity',
        'power_display_color',
        'loading_intensity',
        'display_color',
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._snapshot: Snapshot | None = None
        self.tuple_index: tuple[int] | tuple[int, int] = (
            (self.index,) if isinstance(self.index, int) else self.index
        )
        self.display_index: str = ':'.join(map(str, self.tuple_index))

    def as_snapshot(self) -> Snapshot:
        self._snapshot = super().as_snapshot()
        self._snapshot.tuple_index = self.tuple_index
        self._snapshot.display_index = self.display_index
        return self._snapshot

    @property
    def snapshot(self) -> Snapshot:
        if self._snapshot is None:
            self._snapshot = self.as_snapshot()
        return self._snapshot

    def mig_devices(self) -> list[MigDevice]:  # type: ignore[override]
        mig_devices = []

        if self.is_mig_mode_enabled():
            for mig_index in range(self.max_mig_device_count()):
                try:
                    mig_device = MigDevice(index=(self.index, mig_index))
                except libnvml.NVMLError:  # noqa: PERF203
                    break
                else:
                    mig_devices.append(mig_device)

        return mig_devices

    fan_speed = ttl_cache(ttl=5.0)(DeviceBase.fan_speed)
    temperature = ttl_cache(ttl=5.0)(DeviceBase.temperature)
    display_active = ttl_cache(ttl=5.0)(DeviceBase.display_active)
    display_mode = ttl_cache(ttl=5.0)(DeviceBase.display_mode)
    current_driver_model = ttl_cache(ttl=5.0)(DeviceBase.current_driver_model)
    persistence_mode = ttl_cache(ttl=5.0)(DeviceBase.persistence_mode)
    performance_state = ttl_cache(ttl=5.0)(DeviceBase.performance_state)
    total_volatile_uncorrected_ecc_errors = ttl_cache(ttl=5.0)(
        DeviceBase.total_volatile_uncorrected_ecc_errors,
    )
    compute_mode = ttl_cache(ttl=5.0)(DeviceBase.compute_mode)
    mig_mode = ttl_cache(ttl=5.0)(DeviceBase.mig_mode)

    def power_utilization(self) -> float | NaType:  # in percentage
        power_limit = self.power_limit()
        if not libnvml.nvmlCheckReturn(power_limit, int) or power_limit == 0:
            return NA
        power_usage = self.power_usage()
        if not libnvml.nvmlCheckReturn(power_usage, int):
            return NA
        return round(100.0 * power_usage / power_limit, 1)

    def memory_percent_string(self) -> str:  # in percentage
        return utilization2string(self.memory_percent())

    def memory_utilization_string(self) -> str:  # in percentage
        return utilization2string(self.memory_utilization())

    def gpu_utilization_string(self) -> str:  # in percentage
        return utilization2string(self.gpu_utilization())

    def fan_speed_string(self) -> str:  # in percentage
        return utilization2string(self.fan_speed())

    def temperature_string(self) -> str:  # in Celsius
        temperature = self.temperature()
        return f'{temperature}C' if libnvml.nvmlCheckReturn(temperature, int) else NA

    def memory_loading_intensity(self) -> LoadingIntensity:
        return self.loading_intensity_of(self.memory_percent(), type='memory')

    def bandwidth_loading_intensity(self) -> LoadingIntensity:
        return self.loading_intensity_of(self.memory_utilization(), type='memory')

    def gpu_loading_intensity(self) -> LoadingIntensity:
        return self.loading_intensity_of(self.gpu_utilization(), type='gpu')

    def power_loading_intensity(self) -> LoadingIntensity:
        return self.loading_intensity_of(self.power_utilization(), type='gpu')

    def loading_intensity(self) -> LoadingIntensity:
        return max(self.memory_loading_intensity(), self.gpu_loading_intensity())

    def display_color(self) -> str:
        if self.name().startswith('ERROR:'):
            return 'red'
        return self.loading_intensity().color()

    def memory_display_color(self) -> str:
        if self.name().startswith('ERROR:'):
            return 'red'
        return self.memory_loading_intensity().color()

    def bandwidth_display_color(self) -> str:
        if self.name().startswith('ERROR:'):
            return 'red'
        return self.bandwidth_loading_intensity().color()

    def gpu_display_color(self) -> str:
        if self.name().startswith('ERROR:'):
            return 'red'
        return self.gpu_loading_intensity().color()

    def power_display_color(self) -> str:
        if self.name().startswith('ERROR:'):
            return 'red'
        return self.power_loading_intensity().color()

    @staticmethod
    def loading_intensity_of(
        utilization: float | str,
        type: Literal['memory', 'gpu'] = 'memory',  # pylint: disable=redefined-builtin
    ) -> LoadingIntensity:
        thresholds = {
            'memory': Device.MEMORY_UTILIZATION_THRESHOLDS,
            'gpu': Device.GPU_UTILIZATION_THRESHOLDS,
        }[type]
        if utilization is NA:
            return LoadingIntensity.MODERATE
        if isinstance(utilization, str):
            utilization = utilization.replace('%', '')
        utilization = float(utilization)
        if utilization >= thresholds[-1]:
            return LoadingIntensity.HEAVY
        if utilization >= thresholds[0]:
            return LoadingIntensity.MODERATE
        return LoadingIntensity.LIGHT

    @staticmethod
    def color_of(
        utilization: float | str,
        type: Literal['memory', 'gpu'] = 'memory',  # pylint: disable=redefined-builtin
    ) -> str:
        return Device.loading_intensity_of(utilization, type=type).color()


class MigDevice(MigDeviceBase, Device):  # type: ignore[misc]
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._snapshot: Snapshot | None = None
        self.tuple_index: tuple[int] | tuple[int, int] = (
            (self.index,) if isinstance(self.index, int) else self.index
        )
        self.display_index: str = ':'.join(map(str, self.tuple_index))

    def memory_usage(self) -> str:  # string of used memory over total memory (in human readable)
        return f'{self.memory_used_human()} / {self.memory_total_human():>8s}'

    loading_intensity = Device.memory_loading_intensity

    SNAPSHOT_KEYS: ClassVar[list[str]] = [
        'name',
        'memory_used',
        'memory_free',
        'memory_total',
        'memory_used_human',
        'memory_free_human',
        'memory_total_human',
        'memory_percent',
        'memory_usage',
        'bar1_memory_used_human',
        'bar1_memory_percent',
        'gpu_utilization',
        'memory_utilization',
        'total_volatile_uncorrected_ecc_errors',
        'mig_mode',
        'is_mig_device',
        'gpu_instance_id',
        'compute_instance_id',
        'memory_percent_string',
        'memory_utilization_string',
        'gpu_utilization_string',
        'memory_loading_intensity',
        'memory_display_color',
        'gpu_loading_intensity',
        'gpu_display_color',
        'loading_intensity',
        'display_color',
    ]



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-function-docstring

from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar

from nvitop.api import NA, host
from nvitop.api.host import AccessDenied, PsutilError


__all__ = [
    'AccessDenied',
    'PsutilError',
    'cpu_percent',
    'getuser',
    'hostname',
    'load_average',
    'reverse_ppid_map',
    'swap_memory',
    'uptime',
    'virtual_memory',
]


if TYPE_CHECKING:
    from collections.abc import Callable
    from typing_extensions import ParamSpec  # Python 3.10+

    _P = ParamSpec('_P')
    _T = TypeVar('_T')


def ignore_error(*, fallback: Any) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    """Ignore errors in the function."""

    def wrapper(func: Callable[_P, _T]) -> Callable[_P, _T]:
        def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            try:
                return func(*args, **kwargs)
            except Exception:  # noqa: BLE001 # pylint: disable=broad-exception-caught
                return fallback

        return wrapped

    return wrapper


class VirtualMemory(NamedTuple):  # pylint: disable=missing-class-docstring
    total: int = NA  # type: ignore[assignment]
    available: int = NA  # type: ignore[assignment]
    percent: int = NA  # type: ignore[assignment]
    used: int = NA  # type: ignore[assignment]
    free: int = NA  # type: ignore[assignment]


@ignore_error(fallback=VirtualMemory())
def virtual_memory() -> VirtualMemory:
    vm = host.virtual_memory()
    return VirtualMemory(
        total=vm.total,
        available=vm.available,
        percent=vm.percent,
        used=vm.used,
        free=vm.free,
    )


class SwapMemory(NamedTuple):  # pylint: disable=missing-class-docstring
    total: int = NA  # type: ignore[assignment]
    used: int = NA  # type: ignore[assignment]
    free: int = NA  # type: ignore[assignment]
    percent: float = NA  # type: ignore[assignment]


@ignore_error(fallback=SwapMemory())
def swap_memory() -> SwapMemory:
    sm = host.swap_memory()
    return SwapMemory(
        total=sm.total,
        used=sm.used,
        free=sm.free,
        percent=sm.percent,
    )


@ignore_error(fallback=(NA, NA, NA))
def load_average() -> tuple[float, float, float]:
    la = host.load_average()
    if la is None:
        return (NA, NA, NA)  # type: ignore[unreachable]
    return la


if TYPE_CHECKING:
    from nvitop.api.host import cpu_percent, getuser, hostname, reverse_ppid_map, uptime
else:
    cpu_percent = ignore_error(fallback=NA)(host.cpu_percent)
    getuser = ignore_error(fallback=NA)(host.getuser)
    hostname = ignore_error(fallback=NA)(host.hostname)
    uptime = ignore_error(fallback=NA)(host.uptime)
    reverse_ppid_map = ignore_error(fallback=MappingProxyType({}))(host.reverse_ppid_map)



#!/usr/bin/env python3

# To install `nvitop` with specific version of `nvidia-ml-py`, use:
#
#   pip install nvidia-ml-py==xx.yyy.zz nvitop
#
# or
#
#   pip install 'nvitop[cudaXX]'
#

"""Setup script for ``nvitop``."""

from __future__ import annotations

import contextlib
import re
import sys
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import TYPE_CHECKING

from setuptools import setup


if TYPE_CHECKING:
    from collections.abc import Generator
    from types import ModuleType


HERE = Path(__file__).absolute().parent


@contextlib.contextmanager
def vcs_version(name: str, path: Path | str) -> Generator[ModuleType]:
    """Context manager to update version string in a version module."""
    path = Path(path).absolute()
    assert path.is_file()
    module_spec = spec_from_file_location(name=name, location=path)
    assert module_spec is not None
    assert module_spec.loader is not None
    module = sys.modules.get(name)
    if module is None:
        module = module_from_spec(module_spec)
        sys.modules[name] = module
    module_spec.loader.exec_module(module)

    if module.__release__:
        yield module
        return

    content = None
    try:
        try:
            content = path.read_text(encoding='utf-8')
            path.write_text(
                data=re.sub(
                    r"""__version__\s*=\s*('[^']+'|"[^"]+")""",
                    f'__version__ = {module.__version__!r}',
                    string=content,
                ),
                encoding='utf-8',
            )
        except OSError:
            content = None

        yield module
    finally:
        if content is not None:
            with path.open(mode='wt', encoding='utf-8', newline='') as file:
                file.write(content)


if __name__ == '__main__':
    extra_requirements = {
        'lint': [
            'ruff',
            'pylint[spelling]',
            'xdoctest',
            'mypy',
            'typing-extensions',
            'pre-commit',
        ],
        'cuda10': ['nvidia-ml-py == 11.450.51'],
    }

    with vcs_version(name='nvitop.version', path=HERE / 'nvitop' / 'version.py') as version:
        for pynvml_major in sorted(
            {int(pynvml.partition('.')[0]) for pynvml in version.PYNVML_VERSION_CANDIDATES},
        ):
            pynvml_range = [
                pynvml
                for pynvml in version.PYNVML_VERSION_CANDIDATES
                if pynvml.startswith(f'{pynvml_major}.')
            ]
            if len(pynvml_range) == 1:
                extra_requirements[f'cuda{pynvml_major}'] = [
                    f'nvidia-ml-py == {pynvml_range[0]}',
                ]
            elif len(pynvml_range) >= 2:
                extra_requirements[f'cuda{pynvml_major}'] = [
                    f'nvidia-ml-py >= {pynvml_range[0]}, <= {pynvml_range[-1]}',
                ]

        setup(
            name='nvitop',
            version=version.__version__,
            extras_require=extra_requirements,
        )



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from nvitop.tui.library import DisplayableContainer, Selection


if TYPE_CHECKING:
    from nvitop.tui.tui import TUI


__all__ = ['BaseScreen', 'BaseSelectableScreen']


class BaseScreen(DisplayableContainer):
    """Base class for all screens."""

    root: TUI
    parent: TUI

    NAME: ClassVar[str]


class BaseSelectableScreen(BaseScreen):
    """Base class for all selectable screens."""

    selection: Selection
